#*****************************************************************************#
#**
#**  CASAS XMPP Python Library
#**
#**    Brian L Thomas, 2010
#**
#** Tools by the Center for Advanced Studies in Adaptive Systems at
#**  the School of Electrical Engineering and Computer Science at
#**  Washington State University
#** 
#** Copyright Washington State University, 2012
#** Copyright Brian L. Thomas, 2012
#** 
#** Licenced under the Creative Commons (http://creativecommons.org):
#** Attribution-NonCommercial-ShareAlike 3.0 Unported
#** 
#** If this code is used for public research, any resulting publications need
#** to cite work done by Brian L. Thomas at the Center for Advanced Study of 
#** Adaptive Systems (CASAS) at Washington State University.
#** 
#** Contact: Brian L. Thomas (brian.thomas@email.wsu.edu)
#** Contact: Diane J. Cook (cook@eecs.wsu.edu)
#*****************************************************************************#
import sys

if "pygtk" in sys.modules:
    from twisted.internet import gtk2reactor
    gtk2reactor.install()

from twisted.words.protocols.jabber import client, jid, xmlstream
from twisted.words.xish import domish
from twisted.internet import reactor

import datetime
import re
import copy
import time
import signal
import uuid
import xml.dom.minidom
import xml.etree.ElementTree as ETree
import xml.parsers.expat

def getText(nodelist):
    rc = ""
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            rc = rc + node.data
    return str(rc)

def getData(node, message):
    result = ""
    try:
        dom = xml.dom.minidom.parseString(message)
        xmsg = dom.getElementsByTagName(node)
        if xmsg.length > 0:
            result = getText(xmsg[0].childNodes)
    except xml.parsers.expat.ExpatError:
        result = ""
    return result

def dict_to_xml(oDict):
    if len(oDict) != 1:
        print "ERROR: dict_to_xml() needs dict() of len=1"
        return None
    key = oDict.keys()[0]
    root = ETree.Element(key)
    dict_to_xml_children(oDict[key], key, root)
    return root

def dict_to_xml_children(obj, name, xElem):
    if type(obj) == type(list()):
        for dup in obj:
            if type(dup) == type(dict()):
                elem = ETree.SubElement(xElem, name)
                dict_to_xml_children(dup, name, elem)
            else:
                dict_to_xml_children(dup, name, xElem)
    elif type(obj) == type(dict()):
        for key in obj.keys():
            if type(obj[key]) != type(list()):
                elem = ETree.SubElement(xElem, key)
                dict_to_xml_children(obj[key], key, elem)
            else:
                for dup in obj[key]:
                    elem = ETree.SubElement(xElem, key)
                    dict_to_xml_children(dup, key, elem)
    elif type(obj) == type(str()):
        xElem.text = obj
    return

def xmlstr_to_dict(message):
    try:
        xElem = ETree.fromstring(message)
        newDict = xml_to_dict(xElem, dict())
    except xml.parsers.expat.ExpatError:
        newDict = dict()
    return newDict

def xml_to_dict(xElem, newDict=dict()):
    position = None
    if xElem.tag in newDict:
        if type(newDict[xElem.tag]) != type(list()):
            newDict[xElem.tag] = [newDict[xElem.tag]]
        position = len(newDict[xElem.tag])
        if xElem.text:
            newDict[xElem.tag].append(xElem.text)
        else:
            newDict[xElem.tag].append("")
    else:
        if xElem.text:
            newDict[xElem.tag] = xElem.text
        else:
            newDict[xElem.tag] = dict()
    children = xElem.getchildren()
    if children:
        if position == None:
            newDict[xElem.tag] = dict()
        else:
            newDict[xElem.tag][position] = dict()
        for child in children:
            if position == None:
                newDict[xElem.tag] = xml_to_dict(child, newDict[xElem.tag])
            else:
                newDict[xElem.tag][position] = xml_to_dict(child, newDict[xElem.tag][position])
    if newDict[xElem.tag] == dict():
        newDict[xElem.tag] = ""
    return newDict


class Data:
    def clean(self, msg):
        msg = re.sub('_#!#_', '', msg)
        return msg
    
    def get(self, obj, keys):
        val = ""
        if len(keys) > 1:
            if keys[0] in obj:
                val = self.get(obj[keys[0]], keys[1:])
        elif len(keys) > 0:
            if keys[0] in obj:
                val = obj[keys[0]]
        return val


class Event(Data):
    by = ""
    category = ""
    stamp = ""
    location = ""
    message = ""
    serial = ""
    mtype = ""
    uuid = ""
    def __init__(self, msg):
        data = xmlstr_to_dict(msg)
        if data != dict():
            self.build_event(data)
        return
    
    def build_event(self, data):
        self.by = self.clean(self.get(data, ['publish','data','event','by']))
        self.category = self.clean(self.get(data, ['publish','data','event','category']))
        self.stamp = self.clean(self.get(data, ['publish','data','event','epoch']))
        self.location = self.clean(self.get(data, ['publish','data','event','location']))
        self.message = self.clean(self.get(data, ['publish','data','event','message']))
        self.serial = self.clean(self.get(data, ['publish','data','event','serial']))
        self.mtype = self.clean(self.get(data, ['publish','data','event','type']))
        self.uuid = self.clean(self.get(data, ['publish','data','event','uuid']))
        if self.by == "":
            self.by = "unknown"
        if self.category == "":
            self.category = "unknown"
        if self.stamp == "":
            self.stamp = datetime.datetime.now()
        else:
            self.stamp = datetime.datetime.fromtimestamp(float(self.stamp))
        if self.location == "":
            self.location = "unknown"
        if self.message == "":
            self.message = "unknown"
        if self.serial == "":
            self.serial = "unknown"
        if self.mtype == "":
            self.mtype = "unknown"
        if self.uuid == "":
            self.uuid = str(uuid.uuid4().hex)
        return
    
    def get_scribe_event(self):
        msg = "EVENT"
        msg += "_#!#_%s" % self.uuid
        msg += "_#!#_%s" % str(self.stamp)
        msg += "_#!#_%s" % self.serial
        msg += "_#!#_%s" % self.location
        msg += "_#!#_%s" % self.message
        msg += "_#!#_%s" % self.by
        msg += "_#!#_%s" % self.category
        msg += "_#!#_%s" % self.mtype
        return msg


class Tag(Event):
    createdBy = ""
    labelName = ""
    dsetName = ""
    expirName = ""
    def __init__(self, msg):
        data = xmlstr_to_dict(msg)
        if data != dict():
            self.build_event(data)
            self.build_tag(data)
        return
    
    def build_tag(self, data):
        self.createdBy = self.clean(self.get(data, ['publish','data','tag','created_by']))
        self.labelName = self.clean(self.get(data, ['publish','data','tag','label','name']))
        self.labelValue = self.clean(self.get(data, ['publish','data','tag','label','value']))
        self.dsetName = self.clean(self.get(data, ['publish','data','tag','dataset','name']))
        self.expirName = self.clean(self.get(data, ['publish','data','tag','experiment','name']))
        if self.createdBy == "":
            self.createdBy = "unknown"
        if self.dsetName == "":
            self.dsetName = "unknown"
        if self.expirName == "":
            self.expirName = "unknown"
        return
    
    def get_scribe_tag(self):
        msg = "TAG"
        msg += "_#!#_%s" % self.createdBy
        msg += "_#!#_%s" % self.labelName
        msg += "_#!#_%s" % self.labelValue
        msg += "_#!#_%s" % self.dsetName
        msg += "_#!#_%s" % self.expirName
        msg += "_#!#_%s" % self.uuid
        msg += "_#!#_%s" % str(self.stamp)
        msg += "_#!#_%s" % self.serial
        msg += "_#!#_%s" % self.location
        msg += "_#!#_%s" % self.message
        msg += "_#!#_%s" % self.by
        msg += "_#!#_%s" % self.category
        msg += "_#!#_%s" % self.mtype
        return msg


class Command(Data):
    by = ""
    command = ""
    value = ""
    serial = list()
    location = list()
    replyto = ""
    cid = ""
    stamp = ""
    uuid = ""
    def __init__(self, msg):
        data = xmlstr_to_dict(msg)
        if data != dict():
            self.build_command(data)
        return
    
    def build_command(self, data):
        self.by = self.clean(self.get(data, ['publish','data','control','by']))
        self.command = self.clean(self.get(data, ['publish','data','control','command']))
        self.value = self.clean(self.get(data, ['publish','data','control','value']))
        self.serial = self.clean(self.get(data, ['publish','data','control','serial']))
        self.location = self.clean(self.get(data, ['publish','data','control','location']))
        self.replyto = self.clean(self.get(data, ['publish','data','control','replyto']))
        self.cid = self.clean(self.get(data, ['publish','data','control','cid']))
        self.stamp = self.clean(self.get(data, ['publish','data','control','epoch']))
        self.uuid = self.clean(self.get(data, ['publish','data','control','uuid']))
        if self.stamp == "":
            self.stamp = datetime.datetime.now()
        else:
            self.stamp = datetime.datetime.fromtimestamp(float(self.stamp))
        return
    
    def get_scribe_event(self):
        msg = "EVENT"
        msg += "_#!#_%s" % self.uuid
        msg += "_#!#_%s" % str(self.stamp)
        msg += "_#!#_%s" % self.serial
        msg += "_#!#_%s" % self.location
        msg += "_#!#_%s|%s|%s|%s" % (self.command, self.value, self.replyto, self.cid)
        msg += "_#!#_%s" % self.by
        msg += "_#!#_control"
        msg += "_#!#_control"
        return msg


class Connection:
    def __init__(self, name):
        self.name = name
        self.isAuthenticated = False
        self.manager = None
        self.jid = None
        self.server = None
        self.chans_subscribe = []
        self.chans_publish = []
        self.authd_callback = None
        self.buddy_quit_callback = None
        self.channel_callback = dict()
        self.direct_msg_callback = None
        self.error_callback = None
        self.finish_callback = None
        self.reactor = reactor
        self.debugging = False
        self.waiting_roster = False
        self.roster = None
        return
    
    def connect(self, jidVal=None, password=None):
        myJid = jid.JID(jidVal)
        if self.manager == None:
            self.manager = "manager@%s" % myJid.host
        self.jid = myJid.userhost()
        self.server = str(myJid.host)
        myJid.resource = "onlyone"
        factory = client.basicClientFactory(myJid, password)
        factory.addBootstrap('//event/stream/authd', self.authd)
        self.reactor.connectTCP(myJid.host, 5222, factory)
        if self.finish_callback != None:
            print "XMPP Lib: setting finish_callback"
            self.reactor.addSystemEventTrigger('before', signal.SIGINT, self.finish_callback)
        self.reactor.run()
        return
    
    def disconnect(self):
        if self.isAuthenticated:
            self.unpublish_channel("log")
            self.isAuthenticated = False
            if self.reactor.running:
                self.reactor.stop()
        return
    
    def authd(self, xmlstreamobj):
        print "authenticated"
        self.isAuthenticated = True
        
        xmlstreamobj.addObserver('/presence', self.presence)
        xmlstreamobj.addObserver('/message', self.listen)
        xmlstreamobj.addObserver('/iq', self.handle_iq)
        self.connection = xmlstreamobj
        
        self.set_status("%s is up!" % self.name)
        self.subscribe_buddy(self.manager)
        self.publish_channel("log")
        
        self.waiting_roster = True
        iq = domish.Element(('jabber:client', 'iq'))
        iq['type'] = 'get'
        iq.addElement(('jabber:iq:roster', 'query'))
        self.connection.send(iq)
        
        if self.authd_callback != None:
            self.reactor.callLater(1, self.authd_callback)
        return
    
    def clean_jid(self, dirty):
        clean = jid.JID(str(dirty))
        return clean.userhost()
    
    def callLater(self, seconds, function, *args):
        delayCall = self.reactor.callLater(seconds, function, *args)
        return delayCall
    
    def cancelLater(self, delayCall):
        if delayCall.active():
            delayCall.cancel()
        return
    
    def set_custom_manager(self, manager):
        tmp = str(manager).split('/')
        self.manager = str(tmp[0])
        return
    
    def set_authd_callback(self, cb):
        self.authd_callback = cb
        return
    
    def set_direct_msg_callback(self, cb):
        self.direct_msg_callback = cb
        return
    
    def set_channel_callback(self, channel, cb):
        if channel in self.channel_callback and cb == None:
            del self.channel_callback[channel]
        else:
            self.channel_callback[channel] = cb
        return
    
    def set_error_callback(self, cb):
        self.error_callback = cb
        return
    
    def set_finish_callback(self, cb):
        self.finish_callback = cb
        return
    
    def set_buddy_quit_callback(self, cb):
        self.buddy_quit_callback = cb
        return
    
    def handle_iq(self, elem):
        if self.waiting_roster and elem.query.uri == 'jabber:iq:roster':
            print "got roster!"
            self.waiting_roster = False
            self.handle_roster(elem)
        print "=" * 60
        print elem.toXml()
        return
    
    def handle_roster(self, elem):
        self.roster = dict()
        items = re.findall('<item .*?</item>|<item .*?/>', elem.toXml())
        for x in items:
            jid = re.findall("jid='(.*?)'", x)
            print jid[0]
            if re.search("ask='", x):
                ask = re.findall("ask='(.*?)'", x)
                print "    ask =",ask[0]
            if re.search("subscription='", x):
                subscr = re.findall("subscription='(.*?)'", x)
                print "    subscription =",subscr[0]
        return
    
    def presence(self, elem):
        if elem.hasAttribute('type'):
            presence = domish.Element(('jabber:client', 'presence'))
            presence['to'] = str(elem['from'])
            presence['from'] = str(self.jid)
            if elem['type'] == "subscribe":
                presence['type'] = "subscribed"
                self.connection.send(presence)
                #p2 = domish.Element(('jabber:client','presence'))
                #p2['to'] = str(elem['from'])
                #p2['from'] = str(self.jid)
                #p2['type'] = "subscribe"
                #self.connection.send(p2)
            elif elem['type'] == "unsubscribe":
                presence['type'] = "unsubscribed"
                self.connection.send(presence)
            elif elem['type'] == "unavailable":
                print "%s went byby..." % str(elem['from'])
                if self.buddy_quit_callback != None:
                    self.buddy_quit_callback(str(elem['from']))
        else:
            if self.clean_jid(elem['from']) == self.manager:
                print "manager (online","#"*40
                self.handle_manager_online()
            else:
                print "presence (online)","="*40
                print elem.toXml()
        return
    
    def handle_manager_online(self):
        for x in self.chans_subscribe:
            self.do_pubsub("register", "subscriber", x)
        for x in self.chans_publish:
            self.do_pubsub("register", "publisher", x)
        return
    
    def listen(self, elem):
        isError = False
        if elem.hasAttribute('type'):
            if elem['type'] == "error":
                isError = True
                if self.error_callback != None:
                    self.error_callback(str(elem.toXml()))
                #if re.search("error code='503'",str(elem.toXml())):
                #    print "ERROR: Couldn't send to %s" % str(elem['from'])
        if not isError:
            if self.debugging:
                print "msg from:",str(elem['from'])
                print str(elem.toXml())
            whofrom = self.clean_jid(elem['from'])
            body = getData("body", str(elem.toXml()))
            if body != "":
                if re.search("<error>.*?</error>", body):
                    if self.error_callback != None:
                        self.error_callback(body)
                    else:
                        print body
                elif whofrom == self.manager:
                    channel = getData("channel", body)
                    if channel in self.channel_callback:
                        self.channel_callback[channel](body)
                else:
                    if self.direct_msg_callback != None:
                        self.direct_msg_callback(body, self.clean_jid(elem['from']))
        return
    
    def send(self, message, to):
        if self.isAuthenticated:
            msg = domish.Element((None, 'message'))
            msg['to'] = str(to)
            msg['from'] = str(self.jid)
            msg.addElement('body', content=message)
            self.connection.send(msg)
        return
    
    def publish_data(self, by, mtype, serial, location, message, category, channel="rawevents"):
        msg = "<event><by>%s</by>" % str(by)
        msg += "<type>%s</type>" % str(mtype)
        msg += "<serial>%s</serial>" % str(serial)
        msg += "<location>%s</location>" % str(location)
        msg += "<message>%s</message>" % str(message)
        msg += "<category>%s</category>" % str(category)
        msg += "</event>"
        self.publish(msg, channel)
        return
    
    def publish_command(self, by, command, value, serial, location, replyto="", cid="", channel="control"):
        msg = "<control><by>%s</by>" % str(by)
        msg += "<command>%s</command>" % str(command)
        msg += "<value>%s</value>" % str(value)
        msg += "<serial>%s</serial>" % str(serial)
        msg += "<location>%s</location>" % str(location)
        msg += "<replyto>%s</replyto>" % str(replyto)
        msg += "<cid>%s</cid></control>" % str(cid)
        self.publish(msg, channel)
        return
    
    def publish(self, message, channel="rawevents"):
        newMsg = "<publish><channel>%s</channel><data>" % str(channel)
        newMsg += message
        newMsg += "</data></publish>"
        self.send(newMsg, self.manager)
        return
    
    def replay_event(self, by, mtype, serial, location, message, category, epoch, uuid="", channel="replay"):
        msg = "<event><by>%s</by>" % str(by)
        msg += "<type>%s</type>" % str(mtype)
        msg += "<serial>%s</serial>" % str(serial)
        msg += "<location>%s</location>" % str(location)
        msg += "<message>%s</message>" % str(message)
        msg += "<category>%s</category>" % str(category)
        msg += "<epoch>%s</epoch>" % str(epoch)
        msg += "<uuid>%s</uuid>" % str(uuid)
        msg += "</event>"
        self.publish(msg, channel)
        return
    
    def tag_event(self, created_by, experiment, dataset, label, value="", by="", mtype="", serial="", location="", message="", category="", epoch="", uuid="", channel="tag"):
        msg = "<tag>"
        msg += "<created_by>%s</created_by>" % str(created_by)
        msg += "<label><name>%s</name>" % str(label)
        msg += "<value>%s</value></label>" % str(value)
        msg += "<dataset><name>%s</name></dataset>" % str(dataset)
        msg += "<experiment><name>%s</name></experiment>" % str(experiment)
        msg += "</tag>"
        test_event = by + mtype + serial + location + message + category + epoch
        if test_event != "":
            msg += "<event><by>%s</by>" % str(by)
            msg += "<type>%s</type>" % str(mtype)
            msg += "<serial>%s</serial>" % str(serial)
            msg += "<location>%s</location>" % str(location)
            msg += "<message>%s</message>" % str(message)
            msg += "<category>%s</category>" % str(category)
            msg += "<epoch>%s</epoch>" % str(epoch)
            msg += "<uuid>%s</uuid>" % str(uuid)
            msg += "</event>"
        else:
            if channel == "tag":
                if self.error_callback != None:
                    self.error_callback("Publishing to 'tag' requires <event>!")
                else:
                    print "ERROR: Publishing to 'tag' requires <event>!"
                return
        self.publish(msg, channel)
        return
    
    def subscribe_channel(self, channel):
        if channel not in self.chans_subscribe:
            self.chans_subscribe.append(channel)
            self.do_pubsub("register", "subscriber", channel)
        return
    
    def unsubscribe_channel(self, channel):
        if channel in self.chans_subscribe:
            self.chans_subscribe.remove(channel)
            self.do_pubsub("unregister", "subscriber", channel)
        return
    
    def publish_channel(self, channel):
        if channel not in self.chans_publish:
            self.chans_publish.append(channel)
            self.do_pubsub("register", "publisher", channel)
        return
    
    def unpublish_channel(self, channel):
        if channel in self.chans_publish:
            self.chans_publish.remove(channel)
            self.do_pubsub("unregister", "publisher", channel)
        return
    
    def do_pubsub(self, action, who, channel):
        msg = "<system><do>%s</do><as>%s</as>" % (str(action), str(who))
        msg += "<channel>%s</channel></system>" % str(channel)
        self.send(msg, self.manager)
        return
    
    def subscribe_buddy(self, buddy):
        presence = domish.Element(('jabber:client', 'presence'))
        presence['to'] = str(buddy)
        presence['from'] = str(self.jid)
        presence['type'] = "subscribe"
        self.connection.send(presence)
        return
    
    def unsubscribe_buddy(self, buddy):
        presence = domish.Element(('jabber:client', 'presence'))
        presence['to'] = str(buddy)
        presence['from'] = str(self.jid)
        presence['type'] = "unsubscribe"
        self.connection.send(presence)
        return
    
    def log(self, message):
        self.publish_data(self.name, "system", self.name, "system", 
                          message, "system", "log")
        return
    
    def log_finish(self, message):
        print self.name,"  msg:",message
        return
    
    def debug(self, message):
        self.publish_data(self.name, "system", self.name, "system",
                          message, "system", "debug")
        return
    
    def set_status(self, value):
        status = domish.Element(('jabber:client', 'presence'))
        status.addElement('status', content=value)
        self.connection.send(status)
        return

class LocalConnection(Connection):

    def connect(self, jidVal=None, password=None):
        myJid = jid.JID(jidVal)
        if self.manager == None:
            self.manager = "manager@%s" % myJid.host
        self.jid = myJid.userhost()
        self.server = str(myJid.host)
        myJid.resource = "onlyone"
        if self.finish_callback != None:
            print "XMPP Lib: setting finish_callback"
            self.reactor.addSystemEventTrigger('before', signal.SIGINT, self.finish_callback)
        self.authd()
        self.reactor.run()
        return
    
    def disconnect(self):
        if self.isAuthenticated:
            self.local_file.close()
            self.unpublish_channel("log")
            self.isAuthenticated = False
            if self.reactor.running:
                self.reactor.stop()
        return

    def authd(self):
        print "authenticated"
        self.isAuthenticated = True
        self.connection = None
        self.local_file = open("local.dat", "a")
        
        if self.authd_callback != None:
            self.reactor.callLater(1, self.authd_callback)
        return
    
    def handle_iq(self, elem):
       return
    
    def handle_roster(self, elem):
        return
    
    def presence(self, elem):
        return
    
    def listen(self, elem):
        isError = False
        if elem.hasAttribute('type'):
            if elem['type'] == "error":
                isError = True
                if self.error_callback != None:
                    self.error_callback(str(elem.toXml()))
                #if re.search("error code='503'",str(elem.toXml())):
                #    print "ERROR: Couldn't send to %s" % str(elem['from'])
        if not isError:
            if self.debugging:
                print "msg from:",str(elem['from'])
                print str(elem.toXml())
            whofrom = self.clean_jid(elem['from'])
            body = getData("body", str(elem.toXml()))
            if body != "":
                if re.search("<error>.*?</error>", body):
                    if self.error_callback != None:
                        self.error_callback(body)
                    else:
                        print body
                elif whofrom == self.manager:
                    channel = getData("channel", body)
                    if channel == "raw_events":
                       ev = Event(body)
                       self.local_file.write(ev.get_scribe_event())
                       self.local_file.write("\n")
                    elif channel == "tag":
                       t = Tag(body)
                       self.local_file.write(t.get_scribe_tag())
                       self.local_file.write("\n")
                    elif channel == "control":
                       c = Command(body)
                       self.local_file.write(c.get_scribe_event())
                       self.local_file.write("\n")
                    if channel in self.channel_callback:
                        self.channel_callback[channel](body)
                else:
                    if self.direct_msg_callback != None:
                        self.direct_msg_callback(body, self.clean_jid(elem['from']))
        return

    def handle_publish(self, message, sender):
        if self.debugging:
            print "handle_publish( %s )" % message
        
        newmsg = copy.copy(message)
        channel = getData("channel", message)
        if channel not in ["replay", "debug", "tag"]:
            if channel in ["rawevents", "log", "control"]:
                if re.search("<tag>.*?</tag>", message):
                    self.send_error(sender, "ERROR: Can NOT send <tag> to 'rawevents', 'log', or 'control' channels!")
                    return
            
            if re.search("<epoch>.*?</epoch>", message):
                newmsg = re.sub("<epoch>.*?</epoch>",
                                "<epoch>%f</epoch>" % time.time(),
                                newmsg)
            else:
                newmsg = re.sub("</location>",
                                "</location><epoch>%f</epoch>" % time.time(),
                                newmsg)
            
            if re.search("<uuid>.*?</uuid>", newmsg):
                newmsg = re.sub("<uuid>.*?</uuid>",
                                "<uuid>%s</uuid>" % str(uuid.uuid4().hex),
                                newmsg)
            else:
                newmsg = re.sub("</epoch>",
                                "</epoch><uuid>%s</uuid>" % str(uuid.uuid4().hex),
                                newmsg)
        if channel == "tag":
            if re.search("<tag>.*?</tag>", message):
                if not re.search("<label>.*?</label>", message):
                    self.send_error(sender, "<label> REQUIRED with <tag>!")
                    return
                if not re.search("<event>.*?</event>", message):
                    self.send_error(sender, "<event> REQUIRED for storing <tag>!")
                    return
            else:
                self.send_error(sender, "<tag> REQUIRED when sending to channel 'tag'!")
                return
        
        msg = domish.Element((None, 'message'))
        msg.addElement('body', content=newmsg)
        msg['from'] = self.manager
        self.listen(msg)
        return

    def send(self, message, to=None):
        if self.isAuthenticated:
            self.handle_publish(message, self.name)
        return
    
    def send_error(self, user, message):
        xml = "<error>"
        xml += "<error_from>%s</error_from>" % self.jid
        xml += "<error_to>%s</error_to>" % user
        xml += "<error_msg>%s</error_msg>" % message
        xml += "</error>"
        self.send(xml)
        return
    
    def do_pubsub(self, action, who, channel):
        return

    def subscribe_buddy(self, buddy):
        return
    
    def unsubscribe_buddy(self, buddy):
        return

    def set_status(self, value):
        return
